@@ -1,4 +1,4 @@
-"""pt.py: utilities for pytorch."""
+"""Utilities for pytorch."""
 
 try:
     import torch
@@ -29,19 +29,16 @@ from typing import (
 )
 from unittest.mock import Mock
 
-# pylint: disable=ungrouped-imports
 import numpy as np
 import torch.nn.functional as F
+from corgy.types import KeyValueType, SubClassType
 from torch import nn
 from torch.optim import Adam
 from torch.optim.lr_scheduler import _LRScheduler
-from torch.optim.optimizer import Optimizer  # pylint: disable=no-name-in-module
+from torch.optim.optimizer import Optimizer
 from torch.utils.data import DataLoader, Dataset, TensorDataset
 from tqdm import trange
 
-from shinyutils._argp import ClassType, CommaSeparatedInts, KeyValuePairsType
-from shinyutils._subcls import get_subclasses
-
 try:
     from torch.utils.tensorboard import SummaryWriter
 except ImportError:
@@ -62,7 +59,15 @@ __all__ = ["DEFAULT_DEVICE", "PTOpt", "FCNet", "NNTrainer", "SetTBWriterAction"]
 
 
 class PTOpt:
-    """Wrapper around pytorch optimizer and learning rate scheduler."""
+    """Wrapper around pytorch optimizer and learning rate scheduler.
+
+    Args:
+        weights: Iterable of `Tensor` weights to optimize.
+        optim_cls: `Optimizer` class to use.
+        optim_params: Mapping of parameters to pass to `optim_cls`.
+        lr_sched_cls: `LRScheduler` class to use for scheduling the learning rate.
+        lr_sched_params: Mapping of parameters to pass to `lr_sched_cls`.
+    """
 
     def __init__(
         self,
@@ -71,13 +76,13 @@ class PTOpt:
         optim_params: Mapping[str, Any],
         lr_sched_cls: Optional[Type[_LRScheduler]] = None,
         lr_sched_params: Optional[Mapping[str, Any]] = None,
-    ) -> None:
+    ):
         self.optimizer = optim_cls(weights, **optim_params)
         if lr_sched_cls is None:
             self.lr_sched: Optional[_LRScheduler] = None
         else:
             if lr_sched_params is None:
-                lr_sched_params = dict()
+                lr_sched_params = {}
             self.lr_sched = lr_sched_cls(self.optimizer, **lr_sched_params)
 
     def __repr__(self) -> str:
@@ -86,10 +91,12 @@ class PTOpt:
             r += f"\n{self.lr_sched!r}"
         return r
 
-    def zero_grad(self) -> None:
+    def zero_grad(self):
+        """Call `zero_grad` on underlying optimizer."""
         self.optimizer.zero_grad()
 
-    def step(self) -> None:
+    def step(self):
+        """Call `step` on underlying optimizer and lr scheduler (if present)."""
         self.optimizer.step()
         if self.lr_sched is not None:
             self.lr_sched.step()
@@ -98,18 +105,39 @@ class PTOpt:
     def from_args(
         cls, weights: Iterable[torch.Tensor], args: Namespace, arg_prefix: str = ""
     ) -> "PTOpt":
+        """Create `PTOpt` instance from a namespace of arguments.
+
+        Args:
+            weights: Iterable of `torch.Tensor` weights to optimize.
+            args: Namespace of arguments (argument names are as added by
+                `add_parser_args`).
+            arg_prefix: Prefix for argument names (default: `""`).
+        """
         if arg_prefix:
             arg_prefix += "_"
         argvars = vars(args)
         if f"{arg_prefix}lr_sched_cls" not in argvars:
             argvars[f"{arg_prefix}lr_sched_cls"] = None
-            argvars[f"{arg_prefix}lr_sched_params"] = None
+            argvars[f"{arg_prefix}lr_sched_params"] = []
+
+        def _eval_key_val_pairs(key_val_pairs):
+            _d = {}
+            for k, v in key_val_pairs:
+                try:
+                    _d[k] = int(v)
+                except ValueError:
+                    try:
+                        _d[k] = float(v)  # type: ignore
+                    except ValueError:
+                        _d[k] = v
+            return _d
+
         return cls(
             weights,
             argvars[f"{arg_prefix}optim_cls"],
-            argvars[f"{arg_prefix}optim_params"],
+            _eval_key_val_pairs(argvars[f"{arg_prefix}optim_params"]),
             argvars[f"{arg_prefix}lr_sched_cls"],
-            argvars[f"{arg_prefix}lr_sched_params"],
+            _eval_key_val_pairs(argvars[f"{arg_prefix}lr_sched_params"]),
         )
 
     @staticmethod
@@ -121,42 +149,78 @@ class PTOpt:
         default_optim_params: Optional[Mapping[str, Any]] = None,
         add_lr_decay: bool = True,
     ) -> Union[ArgumentParser, _ArgumentGroup]:
-        """Add options to the base parser for pytorch optimizer and lr scheduling."""
+        """Add options to the base parser for pytorch optimizer and lr scheduling.
+
+        Args:
+            base_parser: Argument parser or group to add arguments to.
+            arg_prefix: Prefix for argument names (default: empty string).
+            group_title: Title to use for added options. If `None`, arguments will be
+                added to the base parser. Otherwise, options will be added to a group
+                with the given title. A new group will be created if `base_parser` is
+                not a group.
+            default_optim_cls: Default `Optimizer` class (default: `Adam`).
+            default_optim_params: Default set of parameters to pass to the optimizer
+                (default: `None`).
+            add_lr_decay: Whether to add options for lr decay (default: `True`).
+
+        Example::
+
+            >>> arg_parser = ArgumentParser(
+                    add_help=False, formatter_class=corgy.CorgyHelpFormatter
+            )
+            >>> PTOpt.add_parser_args(arg_parser)
+            >>> arg_parser.print_help()
+            options:
+              --optim-cls cls
+                  ({'Adadelta'/'Adagrad'/'Adam'/'AdamW'/'SparseAdam'/'Adamax'/'ASGD'
+                  /'SGD'/'Rprop'/'RMSprop'/'LBFGS'/'Adam'/'AdamW'/'SGD'/'RMSprop'/'R
+                  prop'/'ASGD'/'Adamax'/'Adadelta'} default: <class
+                  'torch.optim.adam.Adam'>)
+              --optim-params [key=val [key=val ...]]
+                  (default: [])
+              --lr-sched-cls cls
+                  ({'LambdaLR'/'MultiplicativeLR'/'StepLR'/'MultiStepLR'/'Exponentia
+                  lLR'/'CosineAnnealingLR'/'CyclicLR'/'CosineAnnealingWarmRestarts'/
+                  'OneCycleLR'/'SWALR'} optional)
+              --lr-sched-params [key=val [key=val ...]]
+                  (default: [])
+        """
         if arg_prefix:
             arg_prefix += "-"
         if group_title is not None:
             base_parser = base_parser.add_argument_group(group_title)
 
+        opt_sub_cls_type = SubClassType(Optimizer)
         base_parser.add_argument(
             f"--{arg_prefix}optim-cls",
-            type=ClassType(Optimizer),
-            choices=get_subclasses(Optimizer),
+            type=opt_sub_cls_type,
+            choices=list(opt_sub_cls_type.choices()),
             required=default_optim_cls is None,
             default=default_optim_cls,
         )
 
         if default_optim_params is None:
-            default_optim_params = dict()
+            default_optim_params = {}
         base_parser.add_argument(
             f"--{arg_prefix}optim-params",
-            type=KeyValuePairsType(),
-            default=default_optim_params,
+            type=KeyValueType(),
+            nargs="*",
+            default=list(default_optim_params.items()),
         )
 
         if not add_lr_decay:
             return base_parser
 
+        lr_sched_sub_cls_type = SubClassType(_LRScheduler)
         base_parser.add_argument(
             f"--{arg_prefix}lr-sched-cls",
-            type=ClassType(_LRScheduler),
-            choices=get_subclasses(_LRScheduler),
+            type=lr_sched_sub_cls_type,
+            choices=list(lr_sched_sub_cls_type.choices()),
             default=None,
         )
 
         base_parser.add_argument(
-            f"--{arg_prefix}lr-sched-params",
-            type=KeyValuePairsType(),
-            default=dict(),
+            f"--{arg_prefix}lr-sched-params", type=KeyValueType(), nargs="*", default=[]
         )
 
         return base_parser
@@ -165,7 +229,32 @@ class PTOpt:
     def add_help(
         base_parser: Union[ArgumentParser, _ArgumentGroup],
         group_title: Optional[str] = "pytorch help",
-    ) -> None:
+    ):
+        """Add parser arguments for help on PyTorch optimizers and lr schedulers.
+
+        Args:
+            base_parser: `ArgumentParser` or `ArgumentGroup` to add options to.
+            group_title: Title for the group of options (default: `pytorch help`).
+                If `group_title` is `None`, the options are added to the base parser,
+                otherwise they are added to a group (a new one is created if
+                `base_parser` is not a group).
+
+        Example::
+
+            >>> arg_parser = ArgumentParser(
+                    add_help=False, formatter_class=corgy.CorgyHelpFormatter
+            )
+            >>> PTOpt.add_help(arg_parser)
+            >>> arg_parser.print_help()
+            pytorch help:
+              --explain-optimizer cls  describe arguments of a torch optimizer
+                                       (optional)
+              --explain-lr-sched cls   describe arguments of a torch lr scheduler
+                                       (optional)
+            >>> arg_parser.parse_args(["--explain-optimizer", "Adamax"])
+            Adamax(params, lr=0.002, betas=(0.9, 0.999), eps=1e-08, weight_decay=0)
+        """
+
         class _ShowHelp(Action):
             def __call__(self, parser, namespace, values, option_string=None):
                 cls_name = values.__name__
@@ -178,20 +267,28 @@ class PTOpt:
 
         base_parser.add_argument(
             "--explain-optimizer",
-            type=ClassType(Optimizer),
+            type=SubClassType(Optimizer),
             action=_ShowHelp,
             help="describe arguments of a torch optimizer",
         )
         base_parser.add_argument(
             "--explain-lr-sched",
-            type=ClassType(_LRScheduler),
+            type=SubClassType(_LRScheduler),
             action=_ShowHelp,
             help="describe arguments of a torch lr scheduler",
         )
 
 
 class FCNet(nn.Module):
-    """Template for a fully connected network."""
+    """Template for a fully connected network.
+
+    Args:
+        in_dim: Number of input features.
+        out_dim: Number of output features.
+        hidden_sizes: List of hidden layer sizes.
+        hidden_act: Activation function for hidden layers (default: `relu`).
+        out_act: Activation function for output layer (default: `None`).
+    """
 
     _ActType = Callable[[torch.Tensor], torch.Tensor]
 
@@ -202,7 +299,7 @@ class FCNet(nn.Module):
         hidden_sizes: List[int],
         hidden_act: _ActType = F.relu,
         out_act: Optional[_ActType] = None,
-    ) -> None:
+    ):
         super().__init__()
         layer_sizes = [in_dim] + hidden_sizes + [out_dim]
         self.layers = nn.ModuleList(
@@ -211,7 +308,7 @@ class FCNet(nn.Module):
         self.hidden_act = hidden_act
         self.out_act = out_act
 
-    def __repr__(self) -> str:
+    def __repr__(self):
         out_act_repr = self.out_act.__name__ if self.out_act is not None else "None"
         return (
             "hidden, output activation: "
@@ -220,6 +317,7 @@ class FCNet(nn.Module):
         )
 
     def forward(self, x: torch.Tensor) -> torch.Tensor:
+        """Propagate input tensor through the network."""
         for layer in self.layers[:-1]:
             x = self.hidden_act(layer(x))
         x = self.layers[-1](x)
@@ -228,7 +326,13 @@ class FCNet(nn.Module):
         return x
 
     @classmethod
-    def from_args(cls, args: Namespace, arg_prefix: str = ""):
+    def from_args(cls, args: Namespace, arg_prefix: str = "") -> "FCNet":
+        """Create `FCNet` instance from a namespace of arguments.
+
+        Args:
+            args: Namespace of arguments, with names as added by `add_parser_args`.
+            arg_prefix: Prefix for argument names (default: `""`).
+        """
         if arg_prefix:
             arg_prefix += "_"
         argvars = vars(args)
@@ -251,10 +355,43 @@ class FCNet(nn.Module):
         default_hidden_act: Optional[_ActType] = F.relu,
         default_out_act: Optional[_ActType] = None,
     ):
-        """Add options to base_parser for building a FCNet object."""
+        """Add options to a parser for building a `FCNet` object.
+
+        Args:
+            base_parser: `ArgumentParser` or `ArgumentGroup` to add options to.
+            arg_prefix: Prefix for argument names (default: `""`).
+            group_title: Title for the group of options If `None` (the default),
+                the options are added to the base parser. Otherwise they are added
+                to a group with the given title (a new one is created if `base_parser`
+                is not a group).
+            default_indim: Default value for `indim` (default: `None`).
+            default_outdim: Default value for `outdim` (default: `None`).
+            default_hidden_sizes: Default value for `hidden_sizes` (default: `None`).
+            default_hidden_act: Default value for hidden activation (default: `relu`).
+            default_out_act: Default value for output activation (default: `None`).
+
+        Example::
+
+            >>> arg_parser = ArgumentParser(
+                    add_help=False, formatter_class=corgy.CorgyHelpFormatter)
+            )
+            >>> FCNet.add_parser_args(arg_parser)
+            >>> arg_parser.print_help()
+            options:
+              --fcnet-indim int
+                  (required)
+              --fcnet-outdim int
+                  (required)
+              --fcnet-hidden-sizes int [int ...]
+                  (required)
+              --fcnet-hidden-act func
+                  (default: <function relu at 0x7faf08f86830>)
+              --fcnet-out-act func
+                  (optional)
+        """
 
         class _Act:
-            metavar = "func"
+            __metavar__ = "func"
 
             def __call__(self, string):
                 try:
@@ -283,7 +420,8 @@ class FCNet(nn.Module):
         )
         base_parser.add_argument(
             f"--{arg_prefix}fcnet-hidden-sizes",
-            type=CommaSeparatedInts(),
+            type=int,
+            nargs="+",
             required=default_hidden_sizes is None,
             default=default_hidden_sizes,
         )
@@ -302,7 +440,16 @@ class FCNet(nn.Module):
 
 
 class NNTrainer:
-    """Helper class for training a model on a dataset."""
+    """Helper class for training a model on a dataset.
+
+    Args:
+        batch_size: Batch size for training.
+        data_load_workers: Number of workers for loading data (default: `0`).
+        shuffle: Whether to shuffle the data (default: `True`).
+        pin_memory: When to pin data to CUDA memory (default: `True`).
+        drop_last: Whether to drop the last incomplete batch (default: `True`).
+        device: Device to use for training (default: `cuda` if available, else `cpu`).
+    """
 
     def __init__(
         self,
@@ -312,7 +459,7 @@ class NNTrainer:
         pin_memory: bool = True,
         drop_last: bool = True,
         device: torch.device = DEFAULT_DEVICE,
-    ) -> None:
+    ):
         self._batch_size = batch_size
         self._data_load_workers = data_load_workers
         self._shuffle = shuffle
@@ -324,18 +471,24 @@ class NNTrainer:
         self._data_loader: DataLoader
 
     @overload
-    def set_dataset(self, value: Dataset) -> None:
+    def set_dataset(self, value: Dataset):
         ...
 
     @overload
-    def set_dataset(self, value: Tuple[torch.Tensor, ...]) -> None:  # type: ignore
+    def set_dataset(self, value: Tuple[torch.Tensor, ...]):
         ...
 
     @overload
-    def set_dataset(self, value: Tuple[np.ndarray, ...]) -> None:  # type: ignore
+    def set_dataset(self, value: Tuple[np.ndarray, ...]):
         ...
 
     def set_dataset(self, value):
+        """Set the training data.
+
+        Args:
+            value: `torch.utils.data.Dataset` instance, or tuple of `torch.Tensor` or
+                `np.ndarray` objects.
+        """
         if isinstance(value, Dataset):
             self._dataset = value
         elif isinstance(value, tuple):
@@ -362,6 +515,15 @@ class NNTrainer:
         iters: int,
         pbar_desc: str = "Training",
     ):
+        """Train a model.
+
+        Args:
+            model: Model (`nn.Module` instance) to train.
+            opt: `PTOpt` instance to use for optimizing.
+            loss_fn: Loss function mapping input tensors to a loss tensor.
+            iters: Number of iterations to train for.
+            pbar_desc: Description for progress bar (default: `Training`).
+        """
         if self._dataset is None:
             raise RuntimeError("dataset not set: call set_dataset before train")
         bat_iter = iter(self._data_loader)
@@ -388,7 +550,23 @@ class NNTrainer:
 
 
 class SetTBWriterAction(Action):
-    """Set attribute in argparse namespace holding tensorboard writer."""
+    """`argparse.Action` to set the `tb_writer` attribute.
+
+    The attribute (configurable via `SetTBWriterAction.attr`) is set to a
+    `SummaryWriter`, or a `Mock` of the class (if called without any value).
+
+    Usage::
+
+        arg_parser = ArgumentParser()
+        arg_parser.add_argument(
+            "--tb-dir",
+            type=str,
+            action=SetTBWriterAction,
+        )
+        arg_parser.set_defaults(tb_writer=Mock(SummaryWriter))
+        args = arg_parser.parse_args(["--tb-dir", "tmp/tb"])
+        args.tb_writer  # `SummaryWriter` instance
+    """
 
     attr = "tb_writer"
 
@@ -401,7 +579,6 @@ class SetTBWriterAction(Action):
 
 
 def _better_lr_sched_repr(lr_sched: _LRScheduler):
-    """Replace default __repr__ in _LRScheduler."""
     return (
         lr_sched.__class__.__name__
         + "(\n    "
@@ -415,13 +592,3 @@ def _better_lr_sched_repr(lr_sched: _LRScheduler):
 
 
 setattr(_LRScheduler, "__repr__", _better_lr_sched_repr)
-
-# if ENABLE_TB:
-#     shiny_arg_parser.add_argument(
-#         "--tb-dir",
-#         type=OutputDirectoryType(),
-#         help="tensorboard log directory",
-#         default=None,
-#         action=SetTBWriterAction,
-#     )
-#     shiny_arg_parser.set_defaults(**{SetTBWriterAction.attr: Mock(SummaryWriter)})
